﻿// Finaquant Analytics - http://finaquant.com/
// Copyright © Tunc Ali Kütükcüoglu - Finaquant Analytics GmbH
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Data;
using System.Text.RegularExpressions;
using FinaquantProtos;
using Excel = Microsoft.Office.Interop.Excel;
using System.Windows.Forms;

namespace FinaquantProtosStarter
{
    class Applications
    {

        /// <summary>Demonstration function for Allocations and Distributions
        /// 
        /// </summary>
        /// <remarks>
        /// Copyright © Tunc Ali Kütükcüoglu - Finaquant Analytics GmbH
        /// Finaquant Analytics - http://finaquant.com/
        /// </remarks>
        public static void Allocation_Distribution_demo()
        {
            #region "create tables"

            // define metadata
            MetaData md = MetaData.CreateEmptyMasterData();
            MetaData.AddNewField(md, "department", FieldType.TextAttribute);
            MetaData.AddNewField(md, "country", FieldType.TextAttribute);
            MetaData.AddNewField(md, "year", FieldType.IntegerAttribute);
            MetaData.AddNewField(md, "team", FieldType.TextAttribute);
            MetaData.AddNewField(md, "person", FieldType.TextAttribute);
            MetaData.AddNewField(md, "costs", FieldType.KeyFigure);
            MetaData.AddNewField(md, "costs_key", FieldType.KeyFigure);
            MetaData.AddNewField(md, "costs_distributed", FieldType.KeyFigure);
            MetaData.AddNewField(md, "bonus", FieldType.KeyFigure);
            MetaData.AddNewField(md, "bonus_key", FieldType.KeyFigure);
            MetaData.AddNewField(md, "bonus_distributed", FieldType.KeyFigure);
            MetaData.AddNewField(md, "key_sum", FieldType.KeyFigure);
            MetaData.AddNewField(md, "profitcenter_source", FieldType.TextAttribute);
            MetaData.AddNewField(md, "profitcenter_target", FieldType.TextAttribute);
            MetaData.AddNewField(md, "profit", FieldType.KeyFigure);
            MetaData.AddNewField(md, "profit_key", FieldType.KeyFigure);
            MetaData.AddNewField(md, "profit_distributed", FieldType.KeyFigure);

            // cost table
            var CostTable1Fields = TableFields.CreateEmptyTableFields(md);
            TableFields.AddNewField(CostTable1Fields, "department");
            TableFields.AddNewField(CostTable1Fields, "costs");

            MatrixTable CostTable = MatrixTable.CreateTableWithElements_A(CostTable1Fields,
                "Sales", 1500.0,
                "Production", 5000.0,
                "Contolling", 680.0,
                "Administration", 800.0
                );

            // Cost key table
            var KeyTable1Fields = TableFields.CreateEmptyTableFields(md);
            TableFields.AddNewField(KeyTable1Fields, "department");
            TableFields.AddNewField(KeyTable1Fields, "team");
            TableFields.AddNewField(KeyTable1Fields, "person");
            TableFields.AddNewField(KeyTable1Fields, "costs_key");

            MatrixTable CostKeyTable = MatrixTable.CreateTableWithElements_A(KeyTable1Fields,
                "Sales", "Team A", "Tom", 2.0,
                "Sales", "Team A", "Jim", 6.0,
                "Sales", "Team B", "Susan", 3.0,
                "Sales", "Team B", "Agnes", 4.0,
                "Production", "Team C", "Hakan", 1.0,
                "Production", "Team C", "Timur", 2.0,
                "Production", "Team D", "Jang", 3.0,
                "Production", "Team D", "Moris", 4.0
                );

            // Bonus table
            var BonusTableFields = TableFields.CreateEmptyTableFields(md);
            TableFields.AddNewField(BonusTableFields, "department");
            TableFields.AddNewField(BonusTableFields, "country");
            TableFields.AddNewField(BonusTableFields, "bonus");

            MatrixTable BonusTable = MatrixTable.CreateTableWithElements_A(BonusTableFields,
                "Sales", "Bolivia", 1500.0,
                "Sales", "Peru", 1200.0,
                "Production", "Bolivia", 2200.0,
                "Production", "Peru", 3300.0
                );

            // Bonus key table
            var BonusKeyTableFields = TableFields.CreateEmptyTableFields(md);
            TableFields.AddNewField(BonusKeyTableFields, "department");
            TableFields.AddNewField(BonusKeyTableFields, "country");
            TableFields.AddNewField(BonusKeyTableFields, "team");
            TableFields.AddNewField(BonusKeyTableFields, "year");
            TableFields.AddNewField(BonusKeyTableFields, "bonus_key");

            MatrixTable BonusKeyTable = MatrixTable.CreateTableWithElements_A(BonusKeyTableFields,
                "Sales", "Bolivia", "Team A", 2013, 3.0,
                "Sales", "Bolivia", "Team B", 2013, 4.0,
                "Sales", "Bolivia", "Team A", 2014, 1.0,
                "Sales", "Bolivia", "Team B", 2014, 2.0,
                "Sales", "Peru", "Team C", 2013, 4.0,
                "Sales", "Peru", "Team D", 2013, 6.0,
                "Sales", "Peru", "Team C", 2014, 1.0,
                "Sales", "Peru", "Team D", 2014, 1.0,
                "Production", "Bolivia", "Team A", 2013, 4.0,
                "Production", "Bolivia", "Team B", 2013, 4.0,
                "Production", "Bolivia", "Team A", 2014, 1.0,
                "Production", "Bolivia", "Team B", 2014, 2.0,
                "Production", "Peru", "Team C", 2013, 5.0,
                "Production", "Peru", "Team D", 2013, 3.0,
                "Production", "Peru", "Team C", 2014, 1.0,
                "Production", "Peru", "Team D", 2014, 2.0
                );

            // profit table
            var ProfitTableFields = TableFields.CreateEmptyTableFields(md);
            TableFields.AddNewField(ProfitTableFields, "profitcenter_source");
            TableFields.AddNewField(ProfitTableFields, "profit");

            MatrixTable ProfitTable = MatrixTable.CreateTableWithElements_A(ProfitTableFields,
                "Profit Center Main Hub", 12000.0);

            // profit key table
            var ProfitKeyTableFields = TableFields.CreateEmptyTableFields(md);
            TableFields.AddNewField(ProfitKeyTableFields, "profitcenter_source");
            TableFields.AddNewField(ProfitKeyTableFields, "profitcenter_target");
            TableFields.AddNewField(ProfitKeyTableFields, "profit_key");

            MatrixTable ProfitKeyTable = MatrixTable.CreateTableWithElements_A(ProfitKeyTableFields,
                "Profit Center Main Hub", "Profit Center - Computers", 2.0,
                "Profit Center Main Hub", "Profit Center - Office Materials", 1.0,
                "Profit Center Main Hub", "Profit Center - Household goods", 4.0,
                "Profit Center Main Hub", "Profit Center - Digital cameras", 3.0);

            #endregion "create tables"

            //*********************************************************************************
            // Demonstration of distribution in 4 steps
            //*********************************************************************************

            // view table
            MatrixTable.View_MatrixTable(CostTable, "Cost table: Costs per department");

            // view table
            MatrixTable.View_MatrixTable(CostKeyTable, "Key table: Key table for distribution from department to teams and persons");

            // find common attributes of two tables
            TextVector CommonAttributes = TextVector.Intersection(CostKeyTable.ColumnNames,
                CostTable.ColumnNames);
            System.Diagnostics.Debug.WriteLine("Common attributes: " + CommonAttributes);

            // combine tables
            NumVector MatchedRowsTbl1, MatchedRowsTbl2;

            var CombinedTbl1 = MatrixTable.CombineTables(CostKeyTable, CostTable,
                null, null, null, out MatchedRowsTbl1, out MatchedRowsTbl2);

            // filter out unmatched rows
            CombinedTbl1 = MatrixTable.PartitionRow(CombinedTbl1, MatchedRowsTbl1);

            // view table
            MatrixTable.View_MatrixTable(CombinedTbl1, "Step 1: Combined cost and key tables");

            // add key key figure "key_sum" into combined table
            bool IfSuccess;
            string Warnings;

            CombinedTbl1 = MatrixTable.AggregateSelectedKeyFigure_B(CombinedTbl1,
                TextVector.CreateVectorWithElements("department"),
                "costs_key", "key_sum", AggregateOption.nSum, out IfSuccess, out Warnings);

            // view table
            MatrixTable.View_MatrixTable(CombinedTbl1, "Step 2: Key figure key_sum is added to combined table");

            // insert key figure "costs_distributed" into combined table
            CombinedTbl1 = MatrixTable.InsertNewColumn(CombinedTbl1, "costs_distributed", FillAllValue: 0.0);

            // view table
            MatrixTable.View_MatrixTable(CombinedTbl1, "Step 3: New key figure costs_distributed is inserted into to combined table");

            // row-by-row processing with user-defined function
            CombinedTbl1 = MatrixTable.TransformRowsDic(CombinedTbl1, DistributeSourceAmountToTarget,
                "costs_distributed",    // target key figure
                "costs",                // source key figure
                "costs_key",            // key
                "key_sum");             // sum

            // view table
            MatrixTable.View_MatrixTable(CombinedTbl1, "Step 4: Combined table after distribution");

            //*********************************************************************************
            // call function SimpleDistributuion() for the same distribution
            //*********************************************************************************

            MatrixTable CombinedTbl2 = SimpleDistribution(SourceTable: CostTable, KeyTable: CostKeyTable,
                TargetKeyFig: "costs_distributed", KeySumKeyFig: "key_sum");

            // view table
            MatrixTable.View_MatrixTable(CombinedTbl2,
                "Result of SimpleDistributuion(): Output table with distributed amounts");

            //*********************************************************************************
            // Distribution example with bonus and key tables
            //*********************************************************************************

            // view table
            MatrixTable.View_MatrixTable(BonusTable, "Bonus table");

            // view table
            MatrixTable.View_MatrixTable(BonusKeyTable, "Bonus key table");

            // Distribute
            MatrixTable ResultTbl = SimpleDistribution(SourceTable: BonusTable, KeyTable: BonusKeyTable,
                TargetKeyFig: "bonus_distributed", KeySumKeyFig: "key_sum");

            // view table
            MatrixTable.View_MatrixTable(ResultTbl,
                "Result of SimpleDistribution() with bonus and key tables as input: Output table with distributed bonus amounts");

            //*********************************************************************************
            // Allocation example as a special case of Distribution
            //*********************************************************************************

            // view table
            MatrixTable.View_MatrixTable(ProfitTable, "Profit table");

            // view table
            MatrixTable.View_MatrixTable(ProfitKeyTable, "Profit key table");

            // Distribute
            ResultTbl = SimpleDistribution(SourceTable: ProfitTable, KeyTable: ProfitKeyTable,
                TargetKeyFig: "profit_distributed", KeySumKeyFig: "key_sum");

            // view table
            MatrixTable.View_MatrixTable(ResultTbl,
                "Allocation results with profit and key tables as input: Output table with distributed profits");

        }

        /// <summary>
        /// Simple pro-rate distribution; returns table with distributed amounts 
        /// The source amount in SourceTable is disrubuted to target amounts w.r.t. keys (or ratios) given in KeyTable
        /// A typical case: Distributing costs or bonus at department level to teams and persons
        /// </summary>
        /// <param name="SourceTable">Souce table with source amounts to be distributed</param>
        /// <param name="KeyTable">Key table with key amounts (or ratios) used for the distribution</param>
        /// <param name="TargetKeyFig">Name of the target key figure with the resultant distributed amounts</param>
        /// <param name="KeySumKeyFig">Name of the key figure with the sum of key amounts</param>
        /// <returns>Resultant combined table with distributed amounts</returns>
        /// <remarks>
        /// Copyright © Tunc Ali Kütükcüoglu - Finaquant Analytics GmbH
        /// Finaquant Analytics - http://finaquant.com/
        /// </remarks>
        public static MatrixTable SimpleDistribution(MatrixTable SourceTable, MatrixTable KeyTable,
            string TargetKeyFig, string KeySumKeyFig)
        {
            #region "Check input parameters"

            // check if any input table is null or empty
            if (SourceTable == null || SourceTable.IsEmpty || SourceTable.RowCount == 0)
                throw new Exception("SourceTable is null or empty!");
            if (KeyTable == null || KeyTable.IsEmpty || KeyTable.RowCount == 0)
                throw new Exception("KeyTable is null or empty!");

            // check if tables have at least one common attribute
            TextVector CommonAttributes = TextVector.Intersection(SourceTable.ColumnNames, KeyTable.ColumnNames);

            if (CommonAttributes.nLength == 0)
                throw new Exception("No common attribute! There should be at least one common atribute.");

            // check if both tables have exactly one key figure 
            if (SourceTable.KeyFigureFields.nLength != 1)
                throw new Exception("SourceTable does not have exactly one key figure!");
            if (KeyTable.KeyFigureFields.nLength != 1)
                throw new Exception("KeyTable does not have exactly one key figure!");

            // check if both tables share the same MetaData object
            if (SourceTable.metaData != KeyTable.metaData)
                throw new Exception("Both input tables must share the same MetaData!");

            // check key figure names
            if (TargetKeyFig == null || TargetKeyFig == "")
                throw new Exception("TargetKeyFig is null or empty!");
            if (KeySumKeyFig == null || KeySumKeyFig == "")
                throw new Exception("KeySumKeyFig is null or empty!");

            // check if key figures are already defined in MetaData
            MetaData md = SourceTable.metaData;
            if (MetaData.GetFieldType(md, TargetKeyFig) != FieldType.KeyFigure)
                throw new Exception("Key figure TargetKeyFig " + TargetKeyFig + " is not defined in MetaData!");
            if (MetaData.GetFieldType(md, KeySumKeyFig) != FieldType.KeyFigure)
                throw new Exception("Key figure KeySumKeyFig " + KeySumKeyFig + " is not defined in MetaData!");

            // get key figure names
            string SourceKeyfig = SourceTable.KeyFigureFields[0];
            string KeyKeyfig = KeyTable.KeyFigureFields[0];

            // check if key figures are contained by input tables
            if (SourceKeyfig == TargetKeyFig || SourceKeyfig == KeySumKeyFig)
                throw new Exception("Source key figure can not be TargetKeyFig or KeySumKeyFig!");
            if (KeyKeyfig == TargetKeyFig || KeyKeyfig == KeySumKeyFig)
                throw new Exception("Key key figure can not be TargetKeyFig or KeySumKeyFig!");

            #endregion "Check input parameters"

            // Step 1: Combine source and key tables
            NumVector MatchedRowsTbl1, MatchedRowsTbl2;

            var CombinedTbl = MatrixTable.CombineTables(KeyTable, SourceTable,
                null, null, null, out MatchedRowsTbl1, out MatchedRowsTbl2);

            // filter out unmatched rows
            CombinedTbl = MatrixTable.PartitionRow(CombinedTbl, MatchedRowsTbl1);

            // Step 2: Add key figure KeySumKeyFig with aggregated amounts to combined table
            bool IfSuccess;
            string Warnings;

            CombinedTbl = MatrixTable.AggregateSelectedKeyFigure_B(CombinedTbl,
                CommonAttributes,
                KeyKeyfig, KeySumKeyFig, AggregateOption.nSum, out IfSuccess, out Warnings);

            if (!IfSuccess)
                throw new Exception("New key figure " + KeySumKeyFig + " could not be inserted into table! "
                    + Warnings);

            // Step 3: Insert new key figure TargetKeyFig into combined table
            CombinedTbl = MatrixTable.InsertNewColumn(CombinedTbl, TargetKeyFig, FillAllValue: 0.0);

            // Step 4: Calculate distributed amount (TargetKeyFig)

            // Call user-defined function for row-by-row processing (table transformation function)
            CombinedTbl = MatrixTable.TransformRowsDic(CombinedTbl, DistributeSourceAmountToTarget,
                TargetKeyFig,    // target key figure
                SourceKeyfig,                // source key figure
                KeyKeyfig,              // key
                KeySumKeyFig);             // sum

            return CombinedTbl;
        }

        /// <summary>
        /// Calculate distributed amount: 
        /// target_keyfig = source_keyfig * (key_keyfig / sum_keyfig)
        /// User-defined function for table transformation (row-by-row processing)
        /// required for Allocation_Distribution_demo()
        /// </summary>
        /// <param name="TextAttribDic">Dictionary (associative array) for text attribute values</param>
        /// <param name="NumAttribDic">Dictionary (associative array) for numeric attribute values</param>
        /// <param name="KeyFigDic">Dictionary (associative array) for key figure values</param>
        /// <param name="OtherParameters">Ordered string parameters: source_keyfig, target_keyfig, key_keyfig, sum_keyfig</param>
        /// <remarks>
        /// Copyright © Tunc Ali Kütükcüoglu - Finaquant Analytics GmbH
        /// Finaquant Analytics - http://finaquant.com/
        /// </remarks>
        public static void DistributeSourceAmountToTarget(
        ref Dictionary<string, string> TextAttribDic,
        ref Dictionary<string, int> NumAttribDic,
        ref Dictionary<string, double> KeyFigDic,
            params Object[] OtherParameters)
        {
            // read ordered parameters into string variables
            string target_keyfig;
            string source_keyfig;
            string key_keyfig;
            string sum_keyfig;

            try
            {
                target_keyfig = (string)OtherParameters[0];
                source_keyfig = (string)OtherParameters[1];
                key_keyfig = (string)OtherParameters[2];
                sum_keyfig = (string)OtherParameters[3];
            }
            catch (Exception ex)
            {
                throw new Exception("Missing or improper other parameter in DistributeSourceAmountToTarget: " + ex.Message);
            }

            // Calculate distributed costs: 
            KeyFigDic[target_keyfig] = KeyFigDic[source_keyfig] * KeyFigDic[key_keyfig] / KeyFigDic[sum_keyfig];
        }

    }
}
